from __future__ import absolute_import
from .__version__ import __version__
from .tools import tools
from .utKit import utKit
from . import times
from . import commonutils
from . import renderer
from . import files
from .fmultiprocess import fmultiprocess
from . import cl_utils
from . import logs



from __future__ import print_function
from builtins import str
import os
import unittest
import shutil
import unittest
import pickle
import yaml
from fundamentals.utKit import utKit
from fundamentals import tools
from os.path import expanduser
home = expanduser("~")

packageDirectory = utKit("").get_project_root()
settingsFile = packageDirectory + "/test_settings.yaml"
# settingsFile = home + "/.config/soxspipe.recipes/soxspipe.recipes.yaml"
su = tools(
    arguments={"settingsFile": settingsFile},
    docString=__doc__,
    logLevel="DEBUG",
    options_first=False,
    projectName=None,
    defaultSettingsFile=False
)
arguments, settings, log, dbConn = su.setup()

# SETUP AND TEARDOWN FIXTURE FUNCTIONS FOR THE ENTIRE MODULE
moduleDirectory = os.path.dirname(__file__)
utKit = utKit(moduleDirectory)
log, dbConn, pathToInputDir, pathToOutputDir = utKit.setupModule()
utKit.tearDownModule()

try:
    shutil.rmtree(pathToOutputDir)
except:
    pass
# COPY INPUT TO OUTPUT DIR
shutil.copytree(pathToInputDir, pathToOutputDir)

# Recursively create missing directories
if not os.path.exists(pathToOutputDir):
    os.makedirs(pathToOutputDir)

pathToPickleFile = pathToInputDir + "/list-of-dictionaries.p"

listOfDictionaries = pickle.load(open(pathToPickleFile, "rb"))

class test_list_of_dictionaries(unittest.TestCase):

    def test_list_of_dictionaries_function(self):

        from fundamentals.renderer import list_of_dictionaries
        dataSet = list_of_dictionaries(
            log=log,
            listOfDictionaries=listOfDictionaries
        )
        # print(dataSet.table())
        dataSet.table(filepath=pathToOutputDir + "myData.dat")
        # print(dataSet.csv())
        dataSet.csv(filepath=pathToOutputDir + "myData.csv")
        # print(dataSet.mysql(tableName="TNS"))
        dataSet.mysql(tableName="TNS",
                      filepath=pathToOutputDir + "myData.mysql")

        # print(dataSet.json())
        dataSet.json(filepath=pathToOutputDir + "myData.json")
        # print(dataSet.yaml())
        dataSet.yaml(filepath=pathToOutputDir + "myData.yaml")
        # print(dataSet.markdown())
        dataSet.markdown(filepath=pathToOutputDir + "myData.md")

    def test_list_of_dictionaries_function_02(self):

        dataList = [
            {
                "owner": "daisy",
                "pet": "dog",
                "address": "belfast, uk"
            },
            {
                "owner": "john",
                "pet": "snake",
                "address": "the moon"
            },
            {
                "owner": "susan",
                "pet": "crocodile",
                "address": "larne"
            }

        ]

        from fundamentals.renderer import list_of_dictionaries
        dataSet = list_of_dictionaries(
            log=log,
            listOfDictionaries=dataList
        )
        # print(dataSet.table())
        dataSet.table(filepath=pathToOutputDir + "myData.dat")
        # print(dataSet.csv())
        dataSet.csv(filepath=pathToOutputDir + "myData.csv")
        # print(dataSet.mysql(tableName="TNS"))
        dataSet.mysql(tableName="TNS",
                      filepath=pathToOutputDir + "myData.mysql")
        # print(dataSet.json())
        dataSet.json(filepath=pathToOutputDir + "myData.json")
        # print(dataSet.yaml())
        dataSet.yaml(filepath=pathToOutputDir + "myData.yaml")
        # print(dataSet.markdown())
        dataSet.markdown(filepath=pathToOutputDir + "myData.md")

    def test_list_of_dictionaries_function_exception(self):

        from fundamentals.renderer import list_of_dictionaries
        try:
            this = list_of_dictionaries(
                log=log,
                listOfDictionaries=listOfDictionaries,
                fakeKey="break the code"
            )
            this.get()
            assert False
        except Exception as e:
            assert True
            # print(str(e))

        # x-print-testpage-for-pessto-marshall-web-object

    # x-class-to-test-named-worker-function



"""
*Tools for working with files and folders*
"""
from __future__ import absolute_import
from .recursive_directory_listing import recursive_directory_listing
from .fileChunker import fileChunker
from .list_of_dictionaries_to_mysql_inserts import list_of_dictionaries_to_mysql_inserts
from .tag import tag



#!/usr/local/bin/python
# encoding: utf-8
"""
*Render a python list of dictionaries in various list and markup formats*

:Author:
    David Young
"""
from builtins import str
from builtins import range
from builtins import object
import sys
import os
import io
import re
import codecs
import copy
import json
import yaml
from decimal import Decimal
from datetime import datetime
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import convert_dictionary_to_mysql_table


class list_of_dictionaries(object):
    """
    *The dataset object is a list of python dictionaries. Using this class, the data can be rendered as various list and markup formats*

    **Key Arguments**

    - ``log`` -- logger
    - ``listOfDictionaries`` -- the list of dictionaries to render
    - ``reDatetime`` -- a pre-compiled datetime regex. Default *False*fss 


    **Usage**

    To initialise the dataset object:

    ```python
    dataList = [
        {
            "owner": "daisy",
            "pet": "dog",
            "address": "belfast, uk"
        },
        {
            "owner": "john",
            "pet": "snake",
            "address": "the moon"
        },
        {
            "owner": "susan",
            "pet": "crocodile",
            "address": "larne"
        }

    ]

    from fundamentals.renderer import list_of_dictionaries
    dataSet = list_of_dictionaries(
        log=log,
        listOfDictionaries=dataList
    )
    ```

    """

    def __init__(
            self,
            log,
            listOfDictionaries,
            reDatetime=False
    ):
        self.log = log
        self.log.debug("instansiating a new 'list_of_dictionaries' object")
        self.listOfDictionaries = listOfDictionaries
        self.reDatetime = reDatetime

        return None

    @property
    def list(
            self):
        """*Returns the original list of dictionaries*

        **Usage**

        dataSet.list

        """
        return self.listOfDictionaries

    def csv(
        self,
        filepath=None
    ):
        """*Render the data in CSV format*

        **Key Arguments**

        - ``filepath`` -- path to the file to write the csv content to. Default *None*


        **Return**

        - ``renderedData`` -- the data rendered in csv format


        **Usage**

        To render the data set as csv:

        ```python
        print(dataSet.csv())
        ```

        ```text
        owner,pet,address
        daisy,dog,"belfast, uk"
        john,snake,the moon
        susan,crocodile,larne
        ```

        and to save the csv rendering to file:

        ```python
        dataSet.csv("/path/to/myfile.csv")
        ```

        """
        self.log.debug('starting the ``csv`` method')

        renderedData = self._list_of_dictionaries_to_csv("machine")

        if filepath and renderedData != "NO MATCH":

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            writeFile = codecs.open(filepath, encoding='utf-8', mode='w')
            writeFile.write(renderedData)
            writeFile.close()

        self.log.debug('completed the ``csv`` method')
        return renderedData

    def table(
        self,
        filepath=None
    ):
        """*Render the data as a  plain text table*

        **Key Arguments**

        - ``filepath`` -- path to the file to write the table to. Default *None*


        **Return**

        - ``renderedData`` -- the data rendered as a plain text table


        **Usage**

        To render the data set as a plain text table:

        ```python
        print(dataSet.table())
        ```

        ```text
        +--------+------------+--------------+
        | owner  | pet        | address      |
        +========+============+==============+
        | daisy  | dog        | belfast, uk  |
        | john   | snake      | the moon     |
        | susan  | crocodile  | larne        |
        +--------+------------+--------------+
        ```

        and to save the table rendering to file:

        ```python
        dataSet.table("/path/to/myfile.ascii")
        ```

        """
        self.log.debug('starting the ``table`` method')

        self.filepath = filepath
        renderedData = self._list_of_dictionaries_to_csv("human")

        if filepath and len(self.listOfDictionaries):

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            writeFile = codecs.open(filepath, encoding='utf-8', mode='w')
            writeFile.write(renderedData)
            writeFile.close()

        self.log.debug('completed the ``table`` method')
        return renderedData

    def reST(
        self,
        filepath=None
    ):
        """*Render the data as a  resturcturedText table*

        **Key Arguments**

        - ``filepath`` -- path to the file to write the table to. Default *None*


        **Return**

        - ``renderedData`` -- the data rendered as a resturcturedText table


        **Usage**

        To render the data set as a resturcturedText table:

        ```python
        print(dataSet.reST())
        ```

        ```text
        +--------+------------+--------------+
        | owner  | pet        | address      |
        +========+============+==============+
        | daisy  | dog        | belfast, uk  |
        +--------+------------+--------------+
        | john   | snake      | the moon     |
        +--------+------------+--------------+
        | susan  | crocodile  | larne        |
        +--------+------------+--------------+
        ```

        and to save the table rendering to file:

        ```python
        dataSet.reST("/path/to/myfile.rst")
        ```

        """
        self.log.debug('starting the ``table`` method')

        self.filepath = filepath
        renderedData = self._list_of_dictionaries_to_csv("reST")

        if filepath and len(self.listOfDictionaries):

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            writeFile = codecs.open(filepath, encoding='utf-8', mode='w')
            writeFile.write(renderedData)
            writeFile.close()

        self.log.debug('completed the ``table`` method')
        return renderedData

    def markdown(
        self,
        filepath=None
    ):
        """*Render the data as a markdown table*

        **Key Arguments**

        - ``filepath`` -- path to the file to write the markdown to. Default *None*


        **Return**

        - ``renderedData`` -- the data rendered as a markdown table


        **Usage**

        To render the data set as a markdown table:

        ```python
        print(dataSet.markdown())
        ```

        ```markdown
        | owner  | pet        | address      |
        |:-------|:-----------|:-------------|
        | daisy  | dog        | belfast, uk  |
        | john   | snake      | the moon     |
        | susan  | crocodile  | larne        |
        ```

        and to save the markdown table rendering to file:

        ```python
        dataSet.table("/path/to/myfile.md")
        ```

        """
        self.log.debug('starting the ``markdown`` method')

        self.filepath = filepath
        renderedData = self._list_of_dictionaries_to_csv("markdown")

        if filepath and len(self.listOfDictionaries):

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            writeFile = codecs.open(filepath, encoding='utf-8', mode='w')
            writeFile.write(renderedData)
            writeFile.close()

        self.log.debug('completed the ``markdown`` method')
        return renderedData

    def json(
        self,
        filepath=None
    ):
        """*Render the data in json format*

        **Key Arguments**

        - ``filepath`` -- path to the file to write the json content to. Default *None*


        **Return**

        - ``renderedData`` -- the data rendered as json


        **Usage**

        To render the data set as json:

        ```python
        print(dataSet.json())
        ```

        ```json
        [
            {
                "address": "belfast, uk",
                "owner": "daisy",
                "pet": "dog"
            },
            {
                "address": "the moon",
                "owner": "john",
                "pet": "snake"
            },
            {
                "address": "larne",
                "owner": "susan",
                "pet": "crocodile"
            }
        ]
        ```

        and to save the json rendering to file:

        ```python
        dataSet.json("/path/to/myfile.json")
        ```

        """
        self.log.debug('starting the ``json`` method')

        dataCopy = copy.deepcopy(self.listOfDictionaries)
        for d in dataCopy:
            for k, v in list(d.items()):
                if isinstance(v,  datetime):
                    d[k] = v.strftime("%Y%m%dt%H%M%S")

        renderedData = json.dumps(
            dataCopy,
            separators=(',', ': '),
            sort_keys=True,
            indent=4
        )

        if filepath and len(self.listOfDictionaries):

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            writeFile = codecs.open(filepath, encoding='utf-8', mode='w')
            writeFile.write(renderedData)
            writeFile.close()

        self.log.debug('completed the ``json`` method')
        return renderedData

    def yaml(
        self,
        filepath=None
    ):
        """*Render the data in yaml format*

        **Key Arguments**

        - ``filepath`` -- path to the file to write the yaml content to. Default *None*


        **Return**

        - ``renderedData`` -- the data rendered as yaml


        **Usage**

        To render the data set as yaml:

        ```python
        print(dataSet.yaml())
        ```

        ```yaml
        - address: belfast, uk
          owner: daisy
          pet: dog
        - address: the moon
          owner: john
          pet: snake
        - address: larne
          owner: susan
          pet: crocodile
        ```

        and to save the yaml rendering to file:

        ```python
        dataSet.json("/path/to/myfile.yaml")
        ```

        """
        self.log.debug('starting the ``yaml`` method')

        dataCopy = []
        dataCopy[:] = [dict(l) for l in self.listOfDictionaries]
        renderedData = yaml.dump(dataCopy, default_flow_style=False)

        if filepath and len(self.listOfDictionaries):

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            stream = open(filepath, 'w')
            yaml.dump(dataCopy, stream, default_flow_style=False)
            stream.close()

        self.log.debug('completed the ``yaml`` method')
        return renderedData

    def mysql(
        self,
        tableName,
        filepath=None,
        createStatement=None
    ):
        """*Render the dataset as a series of mysql insert statements*

        **Key Arguments**

        - ``tableName`` -- the name of the mysql db table to assign the insert statements to.
        - ``filepath`` -- path to the file to write the mysql inserts content to. Default *None*
        createStatement


        **Return**

        - ``renderedData`` -- the data rendered mysql insert statements (string format)


        **Usage**

        ```python
        print(dataSet.mysql("testing_table"))
        ```

        this output the following:

        ```plain
        INSERT INTO `testing_table` (address,dateCreated,owner,pet) VALUES ("belfast, uk" ,"2016-09-14T16:21:36" ,"daisy" ,"dog")  ON DUPLICATE KEY UPDATE  address="belfast, uk", dateCreated="2016-09-14T16:21:36", owner="daisy", pet="dog" ;
        INSERT INTO `testing_table` (address,dateCreated,owner,pet) VALUES ("the moon" ,"2016-09-14T16:21:36" ,"john" ,"snake")  ON DUPLICATE KEY UPDATE  address="the moon", dateCreated="2016-09-14T16:21:36", owner="john", pet="snake" ;
        INSERT INTO `testing_table` (address,dateCreated,owner,pet) VALUES ("larne" ,"2016-09-14T16:21:36" ,"susan" ,"crocodile")  ON DUPLICATE KEY UPDATE  address="larne", dateCreated="2016-09-14T16:21:36", owner="susan", pet="crocodile" ;
        ```

        To save this rendering to file use:

        ```python
        dataSet.mysql("testing_table", "/path/to/myfile.sql")
        ```

        """
        self.log.debug('starting the ``mysql`` method')

        import re
        if createStatement and "create table if not exists" not in createStatement.lower():
            regex = re.compile(r'^\s*CREATE TABLE ', re.I | re.S)
            createStatement = regex.sub(
                "CREATE TABLE IF NOT EXISTS ", createStatement)

        renderedData = self._list_of_dictionaries_to_mysql_inserts(
            tableName=tableName,
            createStatement=createStatement
        )

        if filepath and len(self.listOfDictionaries):

            # RECURSIVELY CREATE MISSING DIRECTORIES
            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            writeFile = open(filepath, mode='w')
            writeFile.write(renderedData)
            writeFile.close()

        self.log.debug('completed the ``mysql`` method')
        return renderedData

    def _list_of_dictionaries_to_csv(
            self,
            csvType="human"):
        """Convert a python list of dictionaries to pretty csv output

        **Key Arguments**

        - ``csvType`` -- human, machine or reST


        **Return**

        - ``output`` -- the contents of a CSV file

        """
        self.log.debug(
            'starting the ``_list_of_dictionaries_to_csv`` function')
        import unicodecsv as csv

        if not len(self.listOfDictionaries):
            return "NO MATCH"

        dataCopy = copy.deepcopy(self.listOfDictionaries)

        tableColumnNames = list(dataCopy[0].keys())
        columnWidths = []
        columnWidths[:] = [len(tableColumnNames[i])
                           for i in range(len(tableColumnNames))]

        output = io.BytesIO()
        # setup csv styles
        if csvType == "machine":
            delimiter = ","
        elif csvType in ["human", "markdown"]:
            delimiter = "|"
        elif csvType in ["reST"]:
            delimiter = "|"
        if csvType in ["markdown"]:
            writer = csv.writer(output, delimiter=delimiter,
                                quoting=csv.QUOTE_NONE, doublequote=False, quotechar='"', escapechar="\\", lineterminator="\n")
        else:
            writer = csv.writer(output, dialect='excel', delimiter=delimiter,
                                quotechar='"', quoting=csv.QUOTE_MINIMAL, lineterminator="\n")

        if csvType in ["markdown"]:
            dividerWriter = csv.writer(
                output, delimiter="|", quoting=csv.QUOTE_NONE, doublequote=False, quotechar='"', escapechar="\\", lineterminator="\n")
        else:
            dividerWriter = csv.writer(output, dialect='excel', delimiter="+",
                                       quotechar='"', quoting=csv.QUOTE_MINIMAL, lineterminator="\n")
        # add column names to csv
        header = []
        divider = []
        rstDivider = []
        allRows = []

        # clean up data
        for row in dataCopy:
            for c in tableColumnNames:
                if isinstance(row[c], float) or isinstance(row[c], Decimal):
                    row[c] = "%0.9g" % row[c]
                elif isinstance(row[c], datetime):
                    thisDate = str(row[c])[:10]
                    row[c] = "%(thisDate)s" % locals()

        # set the column widths
        for row in dataCopy:
            for i, c in enumerate(tableColumnNames):
                if len(str(row[c])) > columnWidths[i]:
                    columnWidths[i] = len(str(row[c]))

        # table borders for human readable
        if csvType in ["human", "markdown", "reST"]:
            header.append("")
            divider.append("")
            rstDivider.append("")

        for i, c in enumerate(tableColumnNames):
            if csvType == "machine":
                header.append(c)
            elif csvType in ["human", "markdown", "reST"]:
                header.append(
                    c.ljust(columnWidths[i] + 2).rjust(columnWidths[i] + 3))
                divider.append('-' * (columnWidths[i] + 3))
                rstDivider.append('=' * (columnWidths[i] + 3))

        # table border for human readable
        if csvType in ["human", "markdown", "reST"]:
            header.append("")
            divider.append("")
            rstDivider.append("")

        # fill in the data
        for row in dataCopy:
            thisRow = []
            # table border for human readable
            if csvType in ["human", "markdown", "reST"]:
                thisRow.append("")

            for i, c in enumerate(tableColumnNames):
                if csvType in ["human", "markdown", "reST"]:
                    if row[c] == None:
                        row[c] = ""
                    row[c] = str(str(row[c]).ljust(columnWidths[i] + 2)
                                 .rjust(columnWidths[i] + 3))
                thisRow.append(row[c])
            # table border for human readable
            if csvType in ["human", "markdown", "reST"]:
                thisRow.append("")
            allRows.append(thisRow)
            if csvType in ["reST"]:
                allRows.append(divider)

        if csvType == "machine":
            writer.writerow(header)
        if csvType in ["reST"]:
            dividerWriter.writerow(divider)
            writer.writerow(header)
            dividerWriter.writerow(rstDivider)
        if csvType in ["human"]:
            dividerWriter.writerow(divider)
            writer.writerow(header)
            dividerWriter.writerow(divider)
        elif csvType in ["markdown"]:
            writer.writerow(header)
            dividerWriter.writerow(divider)

        # write out the data
        writer.writerows(allRows)
        # table border for human readable
        if csvType in ["human"]:
            dividerWriter.writerow(divider)

        output = output.getvalue()
        output = output.strip()
        try:
            output = output.decode("UTF-8")
        except:
            output = str(output)

        if csvType in ["markdown"]:
            output = output.replace("|--", "|:-")
        if csvType in ["reST"]:
            output = output.replace("|--", "+--").replace("--|", "--+")

        self.log.debug(
            'completed the ``_list_of_dictionaries_to_csv`` function')

        return output

    def _list_of_dictionaries_to_mysql_inserts(
            self,
            tableName,
            createStatement=None):
        """Convert a python list of dictionaries to pretty csv output

        **Key Arguments**

        - ``tableName`` -- the name of the table to create the insert statements for
        - ``createStatement`` -- add this create statement to the top of the file. Will only be executed if no table of that name exists in database. Default *None*


        **Return**

        - ``output`` -- the mysql insert statements (as a string)

        """
        self.log.debug(
            'completed the ````_list_of_dictionaries_to_mysql_inserts`` function')

        if not len(self.listOfDictionaries):
            return "NO MATCH"

        dataCopy = copy.deepcopy(self.listOfDictionaries)

        if createStatement:
            output = createStatement + "\n"
        else:
            output = ""

        inserts = []

        inserts = []
        inserts[:] = [convert_dictionary_to_mysql_table(log=self.log, dictionary=d, dbTableName=tableName, uniqueKeyList=[
        ], dateModified=False, returnInsertOnly=True, replace=True, batchInserts=False, reDatetime=self.reDatetime) for d in dataCopy]
        output += ";\n".join(inserts) + ";"

        self.log.debug(
            'completed the ``_list_of_dictionaries_to_mysql_inserts`` function')
        return output



from __future__ import print_function
from builtins import str
import os
import unittest
import shutil
import unittest
import yaml
from fundamentals.utKit import utKit
from fundamentals import tools
from os.path import expanduser
home = expanduser("~")

packageDirectory = utKit("").get_project_root()
settingsFile = packageDirectory + "/test_settings.yaml"
# settingsFile = home + "/.config/soxspipe.recipes/soxspipe.recipes.yaml"
su = tools(
    arguments={"settingsFile": settingsFile},
    docString=__doc__,
    logLevel="DEBUG",
    options_first=False,
    projectName=None,
    defaultSettingsFile=False
)
arguments, settings, log, dbConn = su.setup()

# SETUP AND TEARDOWN FIXTURE FUNCTIONS FOR THE ENTIRE MODULE
moduleDirectory = os.path.dirname(__file__)
utKit = utKit(moduleDirectory)
log, dbConn, pathToInputDir, pathToOutputDir = utKit.setupModule()
utKit.tearDownModule()

try:
    shutil.rmtree(pathToOutputDir)
except:
    pass
# COPY INPUT TO OUTPUT DIR
shutil.copytree(pathToInputDir, pathToOutputDir)

# Recursively create missing directories
if not os.path.exists(pathToOutputDir):
    os.makedirs(pathToOutputDir)

# class test__fetch(unittest.TestCase):

#     def test__fetch_function(self):
#         kwargs = {}
#         kwargs["log"] = log
#         kwargs["settings"] = settings
#         # xt-kwarg_key_and_value

#         testObject = _fetch(**kwargs)
#         testObject.get()

# x-print-testpage-for-pessto-marshall-web-object

# x-class-to-test-named-worker-function



"""
*Tools used to perform download of files and HTML pages from the web*
"""
from __future__ import absolute_import
from ._fetch import _fetch
from ._dump_files_to_local_drive import _dump_files_to_local_drive
from .get_now_datetime_filestamp import get_now_datetime_filestamp
from .extract_filename_from_url import extract_filename_from_url
from .append_now_datestamp_to_filename import append_now_datestamp_to_filename
from .multiobject_download import multiobject_download



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a yaml settings file containing database connection details, setup and return the db connector*

:Author:
    David Young
"""
from builtins import str
import sys
import os
import yaml
try:
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass
os.environ['TERM'] = 'vt100'
from fundamentals import tools

def setup_database_connection(
        pathToYamlFile):
    """*Start a database connection using settings in yaml file* 

    Given the location of a YAML dictionary containing database credientials, this function will setup and return the connection*

    **Key Arguments**

    - ``pathToYamlFile`` -- path to the YAML dictionary.
    

    **Return**

    - ``dbConn`` -- connection to the MySQL database.
    

    **Usage**

    The settings file should be in this form, with all keyword values set:

    ```yaml
    db: unit_tests
    host: localhost
    user: utuser
    password: utpass
    ```

    And here's how to generate the connection object:

    ```python
    from fundamentals.mysql import setup_database_connection
    dbConn = setup_database_connection(
        pathToYamlFile=pathToMyYamlFile
    )
    ```
    
    """
    import sys
    import logging
    import pymysql as ms

    # IMPORT THE YAML CONNECTION DICTIONARY
    try:
        logging.info(
            'importing the yaml database connection dictionary from ' + pathToYamlFile)
        stream = open(pathToYamlFile, 'r')
        connDict = yaml.load(stream)
    except:
        logging.critical(
            'could not load the connect dictionary from ' + pathToYamlFile)
        sys.exit(1)
    # ESTABLISH A DB CONNECTION
    try:
        logging.info('connecting to the ' + connDict[
                     'db'] + ' database on ' + connDict['host'])
        dbConn = ms.connect(
            host=connDict['host'],
            user=connDict['user'],
            passwd=connDict['password'],
            db=connDict['db'],
            use_unicode=True,
            charset='utf8',
            local_infile=1,
            client_flag=ms.constants.CLIENT.MULTI_STATEMENTS,
            connect_timeout=36000
        )
        dbConn.autocommit(True)
    except Exception as e:
        logging.critical('could not connect to the ' + connDict['db'] + ' database on ' + connDict['host'] + ' : '
                         + str(e) + '\n')
    return dbConn

# use the tab-trigger below for new function
# xt-def-function



#!/usr/local/bin/python
# encoding: utf-8
"""
*Generate the datetime stamp for filenames*

:Author:
    David Young
"""
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools


def get_now_datetime_filestamp(longTime=False):
    """
    *A datetime stamp to be appended to the end of filenames: 'YYYYMMDDtHHMMSS'*

    **Key Arguments**

    - ``longTime`` -- make time string longer (more change of filenames being unique)

    **Return**

    - ``now`` -- current time and date in filename format

    **Usage**

    ```python
    from fundamentals.download import get_now_datetime_filestamp
    get_now_datetime_filestamp(longTime=False)
    #Out: '20160316t154635'

    get_now_datetime_filestamp(longTime=True)
    #Out: '20160316t154644133638' 
    ```
    """
    from datetime import datetime, date, time
    now = datetime.now()
    if longTime:
        now = now.strftime("%Y%m%dt%H%M%S%f")
    else:
        now = now.strftime("%Y%m%dt%H%M%S")

    return now



#!/usr/local/bin/python
# encoding: utf-8
"""
*Some time functions to be used with logging etc*

:Author:
    David Young
"""
from builtins import str
import sys
import os
os.environ['TERM'] = 'vt100'

def get_now_sql_datetime():
    """
    *A datetime stamp in MySQL format: 'YYYY-MM-DDTHH:MM:SS'*

    **Return**

    - ``now`` -- current time and date in MySQL format

    **Usage**

    ```python
    from fundamentals import times
    now = times.get_now_sql_datetime()
    print now

    # OUT: 2016-03-18T11:08:23 
    ```
    """
    from datetime import datetime, date, time
    now = datetime.now()
    now = now.strftime("%Y-%m-%dT%H:%M:%S")

    return now

def datetime_relative_to_now(date):
    """
    *convert date to a relative datetime (e.g. +15m, +2hr, +1w)*

    **Key Arguments**

    - ``date`` -- absolute date

    **Return**

    - a relative date

    **Usage**

    ```python
    from fundamentals import times
    relTime = times.datetime_relative_to_now(date)
    ```
    """
    from datetime import datetime
    diff = datetime.now() - date
    s = diff.seconds
    if diff.days == 1:
        return ' + 1d'
    elif diff.days > 1:
        return ' +{0}d'.format(diff.days)
    elif s <= 1:
        return ' just now'
    elif s < 60:
        return ' +{0}sec'.format(s)
    elif s < 120:
        return ' +1min'
    elif s < 3600:
        return ' +{0}min'.format(s / 60)
    elif s < 7200:
        return ' +1hr'
    else:
        return ' +{0}hr'.format(s / 3600)

def calculate_time_difference(startDate, endDate):
    """
    *Return the time difference between two dates as a string*

    **Key Arguments**

    - ``startDate`` -- the first date in YYYY-MM-DDTHH:MM:SS format
    - ``endDate`` -- the final date YYYY-MM-DDTHH:MM:SS format

    **Return**

    - ``relTime`` -- the difference between the two dates in Y,M,D,h,m,s (string)

    **Usage**

    ```python
    from fundamentals import times
    diff = times.calculate_time_difference(startDate="2015-10-13 10:02:12", endDate="2017-11-04 16:47:05")
    print diff

    # OUT: 2yrs 22dys 6h 44m 53s
    ```
    """
    from datetime import datetime
    from dateutil import relativedelta

    ################ > VARIABLE SETTINGS ######

    ################ >ACTION(S) ################
    if "T" not in startDate:
        startDate = startDate.strip().replace(" ", "T")
    if "T" not in endDate:
        endDate = endDate.strip().replace(" ", "T")
    startDate = datetime.strptime(startDate, '%Y-%m-%dT%H:%M:%S')
    endDate = datetime.strptime(endDate, '%Y-%m-%dT%H:%M:%S')
    d = relativedelta.relativedelta(endDate, startDate)

    relTime = ""
    if d.years > 0:
        relTime += str(d.years) + "yrs "
    if d.months > 0:
        relTime += str(d.months) + "mths "
    if d.days > 0:
        relTime += str(d.days) + "dys "
    if d.hours > 0:
        relTime += str(d.hours) + "h "
    if d.minutes > 0:
        relTime += str(d.minutes) + "m "
    if d.seconds > 0:
        relTime += str(d.seconds) + "s"
    ###############################

    if relTime == "":
        relTime = "0s"

    return relTime



#!/usr/local/bin/python
# encoding: utf-8
"""
*Logger setup for python projects*

:Author:
    David Young
"""
from builtins import object
import os
import sys
import logging
from logging import handlers

def console_logger(
    level="WARNING"
):
    """
    *Setup and return a console logger*

    **Key Arguments**

    - ``level`` -- the level of logging required
    

    **Return**

    - ``logger`` -- the console logger
    

    **Usage**

    ```python
    from fundamentals import logs
    log = logs.console_logger(
        level="DEBUG"
    )
    log.debug("Testing console logger")
    ```
    
    """
    ## STANDARD LIB ##

    import logging
    import logging.config
    ## THIRD PARTY ##
    import yaml
    try:
        yaml.warnings({'YAMLLoadWarning': False})
    except:
        pass
    ## LOCAL APPLICATION ##

    # SETUP LOGGING
    loggerConfig = """
  version: 1
  formatters:
      console_style:
          format: '* %(asctime)s - %(levelname)s: %(pathname)s:%(funcName)s:%(lineno)d > %(message)s'
          datefmt: '%H:%M:%S'
  handlers:
      console:
          class: logging.StreamHandler
          level: """ + level + """
          formatter: console_style
          stream: ext://sys.stdout
  root:
      level: """ + level + """
      handlers: [console]"""

    logging.config.dictConfig(yaml.load(loggerConfig))
    logger = logging.getLogger(__name__)

    return logger

def setup_dryx_logging(yaml_file):
    """
    *setup dryx style python logging*

    **Key Arguments**

    - ``level`` -- the level of logging required
    

    **Return**

    - ``logger`` -- the console logger
    

    **Usage**

    ```python
    from fundamentals import logs
    log = logs.setup_dryx_logging(
        yaml_file="/Users/Dave/.config/fundamentals/fundamentals.yaml"
    )
    log.error("test error")
    ```

    Here is an example of the settings in the yaml file:

    ```yaml
    version: 1

    logging settings:
        formatters:
            file_style:
                format: '* %(asctime)s - %(name)s - %(levelname)s (%(pathname)s > %(funcName)s > %(lineno)d) - %(message)s  '
                datefmt: '%Y/%m/%d %H:%M:%S'
            console_style:
                format: '* %(asctime)s - %(levelname)s: %(pathname)s:%(funcName)s:%(lineno)d > %(message)s'
                datefmt: '%H:%M:%S'
            html_style:
                format: '<div id="row" class="%(levelname)s"><span class="date">%(asctime)s</span>   <span class="label">file:</span><span class="filename">%(filename)s</span>   <span class="label">method:</span><span class="funcName">%(funcName)s</span>   <span class="label">line#:</span><span class="lineno">%(lineno)d</span> <span class="pathname">%(pathname)s</span>  <div class="right"><span class="message">%(message)s</span><span class="levelname">%(levelname)s</span></div></div>'
                datefmt: '%Y-%m-%d <span class= "time">%H:%M <span class= "seconds">%Ss</span></span>'
        handlers:
            console:
                class: logging.StreamHandler
                level: DEBUG
                formatter: console_style
                stream: ext://sys.stdout
            file:
                class: logging.handlers.GroupWriteRotatingFileHandler
                level: WARNING
                formatter: file_style
                filename: /Users/Dave/.config/fundamentals/fundamentals.log
                mode: w+
                maxBytes: 102400
                backupCount: 1
        root:
            level: WARNING
            handlers: [file,console]
    ```
    
    """
    import logging
    import logging.config
    import yaml
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO
    from os.path import expanduser

    # IMPORT CUSTOM HANDLER THAT ALLOWS GROUP WRITING
    handlers.GroupWriteRotatingFileHandler = GroupWriteRotatingFileHandler

    # GET CONTENT OF YAML FILE AND REPLACE ~ WITH HOME DIRECTORY PATH
    with open(yaml_file) as f:
        content = f.read()
    home = expanduser("~")
    content = content.replace("~/", home + "/")
    stream = StringIO(content)
    yamlContent = yaml.load(stream)
    stream.close()

    # USE THE LOGGING SETTINGS SECTION OF THE SETTINGS DICTIONARY FILE IF THERE IS ONE
    # OTHERWISE ASSUME THE FILE CONTAINS ONLY LOGGING SETTINGS
    if "logging settings" in yamlContent:
        yamlContent = yamlContent["logging settings"]
        yamlContent["version"] = 1

    if "handlers" in yamlContent and "file" in yamlContent["handlers"] and "filename" in yamlContent["handlers"]["file"]:
        loggingDir = os.path.dirname(
            yamlContent["handlers"]["file"]["filename"])
        # Recursively create missing directories
        if not os.path.exists(loggingDir):
            os.makedirs(loggingDir)

    logging.config.dictConfig(yamlContent)
    # SET THE ROOT LOGGER
    logger = logging.getLogger(__name__)

    logging.captureWarnings(True)

    return logger

class GroupWriteRotatingFileHandler(handlers.RotatingFileHandler):
    """
    *rotating file handler for logging*
    """

    def doRollover(self):
        """
        *Override base class method to make the new log file group writable.*
        """
        # Rotate the file first.
        handlers.RotatingFileHandler.doRollover(self)

        # Add group write to the current permissions.
        currMode = os.stat(self.baseFilename).st_mode
        os.chmod(self.baseFilename, currMode | stat.S_IWGRP |
                 stat.S_IRGRP | stat.S_IWOTH | stat.S_IROTH)

class GroupWriteRotatingFileHandler(handlers.RotatingFileHandler):
    """
    *rotating file handler for logging*
    """

    def _open(self):
        prevumask = os.umask(0)
        rtv = logging.handlers.RotatingFileHandler._open(self)
        os.umask(prevumask)
        return rtv

class emptyLogger(object):
    """
    *A fake logger object so user can set ``log=False`` if required*

    **Usage**

    ```python
    if log == False:
        from fundamentals.logs import emptyLogger
        log = emptyLogger()
    ```
    
    """

    def info(self, argu):
        pass

    def error(self, argu):
        pass

    def debug(self, argu):
        pass

    def critical(self, argu):
        pass

    def warning(self, argu):
        pass



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a directory of MySQL scripts, execute the scripts and process the script files according to their success or failure*

Run the following code once to set a login-path for your mysql server:

    mysql_config_editor set --login-path=<uniqueLoginName> --host=localhost --user=<myUsername> --password

This store's your credentials in an encrypted file located at '~/.mylogin.cnf'.
Use `mysql_config_editor print --all` to see all of the login-paths set.

Usage:
    mysqlSucker <pathToDirectory> <loginPath> <databaseName> [-s successRule -f failureRule]

    pathToDirectory       path to the directory containing the sql scripts to run (scripts must have `.sql` extension)
    loginPath             the local-path as set with `mysql_config_editor` (`mysqlSucker -h` for more details)
    databaseName          the name of the database to execute the scripts within

Options:
    -h, --help                                  show this help message
    -s successRule, --success successRule       what to do if script succeeds. Default *None* [None|delete|subFolderName]
    -f failureRule, --failure failureRule       what to do if script fails. Default *None* [None|delete|subFolderName]

:Examples:

    To simply execute the scripts in a directory you can run:

        mysqlSucker /path/to/scriptdir myLoginPath myDatabaseName

    To delete script after thay have executed successfully:

        mysqlSucker /path/to/scriptdir myLoginPath myDatabaseName -s delete

    To move successful script to a `passed` sub-directory of `/path/to/scriptdir` and failed scripts to a `failed` sub-directory

        mysqlSucker /path/to/scriptdir myLoginPath myDatabaseName -s pass -f failed

:Author:
    David Young
"""
from builtins import str
import sys
import os
import time
import collections
import datetime
from subprocess import Popen, PIPE, STDOUT
os.environ['TERM'] = 'vt100'
from fundamentals import tools


def directory_script_runner(
        log,
        pathToScriptDirectory,
        dbConn=False,
        waitForResult=True,
        successRule=None,
        failureRule=None,
        loginPath=False,
        databaseName=False,
        force=True):
    """A function to run all mysql scripts in a given directory (in a modified date order, oldest first) and then act on the script files in accordance with the succcess or failure of their execution

    The function can be run with either with an established database connection (`dbConn`) or with a mysql generated `login-path` name (`loginPath`).

    **with `dbConn`**

    Simply pass the connection `dbConn` established elsewhere in your code.

    **with `loginPath`**

    As it's insecure to pass in mysql database credentials via the command-line, run the following command from the terminal

    ```bash
    mysql_config_editor set --login-path=<uniqueLoginName> --host=localhost --user=<myUsername> --password
    > Enter password:
    ```

    This will store your database credentials in an encrypted file located at '~/.mylogin.cnf'. Use `mysql_config_editor print --all` to see all of the login-paths set.

    The `directory_script_runner` function can work by taking advantage of mysql's `--login-path` argument so not to require knowledge of the database credentials.

    Pass the login-path name via the `loginPath` parameter to use `directory_script_runner` in this manner.

    If both `dbConn` and `loginPath` parameters are given, `dbConn` will be given precedent.

    **Key Arguments**

    - ``log`` -- logger
    - ``pathToScriptDirectory`` -- the path to the directory containing the sql script to be run
    - ``databaseName`` -- the name of the database
    - ``force`` -- force the script to run, skipping over lines with errors, Default *True*
    - ``loginPath`` -- the local-path as set with `mysql_config_editor`
    - ``dbConn`` -- the database connection
    - ``waitForResult`` -- wait for the mysql script to finish execution? If 'False' the MySQL script will run in background (do not wait for completion), or if 'delete' the script will run then delete regardless of success status. Default *True*. [True|False|delete]
    - ``successRule`` -- what to do if script succeeds. Default *None* [None|delete|subFolderName]
    - ``failureRule`` -- what to do if script fails. Default *None* [None|delete|subFolderName]


    **Return**

    - None


    **Usage**

    To run the scripts in the directroy and not act on the script file use something similar to:

    ```python
    from fundamentals.mysql import directory_script_runner
    directory_script_runner(
        log=log,
        pathToScriptDirectory="/path/to/mysql_scripts",
        databaseName="imports",
        loginPath="myLoginDetails"
    )
    ```

    To delete successful scripts and archive failed scripts for later inspection:

    ```python
    from fundamentals.mysql import directory_script_runner
    directory_script_runner(
        log=log,
        pathToScriptDirectory="/path/to/mysql_scripts",
        databaseName="imports",
        loginPath="myLoginDetails",
        successRule="delete",
        failureRule="failed"
    )
    ```

    This creates a folder at `/path/to/mysql_scripts/failed` and moves the failed scripts into that folder.

    Finally to execute the scripts within a directory but not wait for the results to return (much fast but you lose error checking in the MySQL scripts):

    ```python
    from fundamentals.mysql import directory_script_runner
    directory_script_runner(
        log=log,
        pathToScriptDirectory="/path/to/mysql_scripts",
        databaseName="imports",
        loginPath="myLoginDetails",
        waitForResults=False
    )
    ```

    Setting ``waitForResults`` = 'delete' will trash the script once it has run (or failed ... be very careful!)

    """
    log.debug('starting the ``directory_script_runner`` function')

    # COMPILE A DICTIONARY OF SCRIPTS / MODIFIED TIMES
    scriptList = {}
    for d in os.listdir(pathToScriptDirectory):
        filePath = os.path.join(pathToScriptDirectory, d)
        filename = os.path.basename(filePath)
        extension = filePath.split(".")[-1]
        if os.path.isfile(filePath) and extension == "sql":
            modified = datetime.datetime.strptime(time.ctime(
                os.path.getmtime(filePath)), "%a %b %d %H:%M:%S %Y")
            scriptList[str(modified) + filename] = filePath

    # ORDER THE DICTIONARY BY MODIFIED TIME - OLDEST FIRST
    scriptList = collections.OrderedDict(sorted(scriptList.items()))

    # FORCE MYSQL SCRIPT?
    if force:
        force = "--force"
    else:
        force = ""

    if dbConn:
        for k, v in list(scriptList.items()):
            scriptname = os.path.basename(v)
            exception = execute_mysql_script(
                pathToScript=v,
                log=log,
                dbConn=dbConn
            )
            # FAILED SCRIPTS
            if exception:
                log.error(
                    'The script %(scriptname)s failed with the following exception: "%(exception)s"' % locals())
                rule = failureRule
                print(exception)
            else:
                rule = successRule
            _process_script_file(
                pathToScriptDirectory=pathToScriptDirectory,
                scriptname=scriptname,
                rule=rule,
                log=log)
    else:
        for k, v in list(scriptList.items()):
            scriptname = os.path.basename(v)
            if waitForResult == True:
                cmd =  """mysql --login-path=%(loginPath)s %(force)s %(databaseName)s < "%(v)s" """ % locals(
                )
                p = Popen(cmd, stdout=PIPE, stderr=PIPE, close_fds=True,
                          env={'PATH': os.getenv('PATH') + ":/usr/local/bin:/usr/bin:/usr/bin:/usr/local/mysql/bin", "MYSQL_TEST_LOGIN_FILE": os.getenv('HOME') + "/.mylogin.cnf"}, shell=True)
                stdout, stderr = p.communicate()

                if len(stderr):
                    log.error(
                        "MySQL Script `%(scriptname)s` Failed: '%(stderr)s'" % locals())
                    rule = failureRule
                else:
                    rule = successRule
                _process_script_file(
                    pathToScriptDirectory=pathToScriptDirectory,
                    scriptname=scriptname,
                    rule=rule,
                    log=log)
            else:
                if waitForResult == "delete":
                    cmd =  """mysql --login-path=%(loginPath)s %(force)s  %(databaseName)s < "%(v)s" > /dev/null 2>&1 & rm "%(v)s" """ % locals()
                else:
                    cmd =  """mysql --login-path=%(loginPath)s %(force)s  %(databaseName)s < "%(v)s" > /dev/null 2>&1 """ % locals()
                p = Popen(cmd, close_fds=True,
                          env={'PATH': os.getenv('PATH') + ":/usr/local/bin:/usr/bin:", "MYSQL_TEST_LOGIN_FILE": os.getenv('HOME') + "/.mylogin.cnf"}, shell=True, stdin=None, stdout=None, stderr=None)

    log.debug('completed the ``directory_script_runner`` function')
    return None


def execute_mysql_script(
        pathToScript,
        dbConn,
        log):
    """*execute a mysql script given its file path and return the success or failure status of the execution*

    **Key Arguments:**

    - `pathToScript` -- path to the sql script to be executed
    - `dbConn` -- mysql database connection
    - `log` -- logger

    **Return:**

    - `exception` -- None or the exception thrown during execution of the sql script. 

    **Usage:**

    To execute a mysql script file located on your local file system:

    ```python
    from fundamentals.mysql import execute_mysql_script
    exception = execute_mysql_script(
        pathToScript="/path/to/my/script.sql",
        log=log,
        dbConn=dbConn
    )

    if exception:
        ...
    ```
    """
    log.debug('starting the ``execute_mysql_script`` function')

    from fundamentals.mysql import writequery

    pathToScript = pathToScript.replace("//", "/")

    import codecs
    pathToReadFile = pathToScript
    try:
        log.debug("attempting to open the file %s" % (pathToReadFile,))
        readFile = codecs.open(pathToReadFile, encoding='utf-8', mode='r')
        thisData = readFile.read()
        readFile.close()
    except IOError as e:
        message = 'could not open the file %s' % (pathToReadFile,)
        log.critical(message)
        raise IOError(message)

    # EXECUTE AND RETURN NONE IF SUCCESSFUL OR THE EXCEPTION IF NOT
    try:
        writequery(
            log=log,
            sqlQuery=thisData,
            dbConn=dbConn
        )
    except Exception as e:
        log.debug('completed the ``execute_mysql_script`` function - failure')
        return e

    log.debug('completed the ``execute_mysql_script`` function - successful')
    return None


def _process_script_file(
        pathToScriptDirectory,
        scriptname,
        rule,
        log):
    """*summary of function*

    **Key Arguments:**

    - `pathToScriptDirectory` -- the path to the directory containing the sql scripts
    - `scriptname` -- name to the sql script to be executed
    - `rule` -- how to process the script. Default *None* [None|delete|subFolderName]
    - `log` -- logger

    **Usage:**

    ```eval_rst
    .. todo::

            add usage info
            create a sublime snippet for usage
    ```

    ```python
    usage code 
    ```           
    """
    log.debug('starting the ``_process_script_file`` function')

    scriptPath = pathToScriptDirectory + "/" + scriptname

    if rule == None or rule == False:
        pass
    elif rule == "delete":
        os.remove(scriptPath)
    elif "/" not in rule:
        moveTo = pathToScriptDirectory + "/" + rule
        # Recursively create missing directories
        if not os.path.exists(moveTo):
            os.makedirs(moveTo)
        moveTo = moveTo + "/" + scriptname
        try:
            log.debug("attempting to rename file %s to %s" %
                      (scriptPath, moveTo))
            os.rename(scriptPath, moveTo)
        except Exception as e:
            log.error(
                "could not rename file %s to %s - failed with this error: %s " % (scriptPath, moveTo, str(e),))

    log.debug('completed the ``_process_script_file`` function')
    return None

# use the tab-trigger below for new function
# xt-def-function


def main(arguments=None):
    """
    The main function used when ``directory_script_runner.py`` is run as a single script from the cl, or when installed as a cl command
    """

    # setup the command-line util settings
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="WARNING",
        options_first=False,
        projectName="fundmentals"
    )
    arguments, settings, log, dbConn = su.setup()

    # UNPACK REMAINING CL ARGUMENTS USING `EXEC` TO SETUP THE VARIABLE NAMES
    # AUTOMATICALLY
    for arg, val in list(arguments.items()):
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        if isinstance(val, ("".__class__, u"".__class__)):
            exec(varname + " = '%s'" % (val,))
        else:
            exec(varname + " = %s" % (val,))
        if arg == "--dbConn":
            dbConn = val
        log.debug('%s = %s' % (varname, val,))

    if successFlag and successFlag.lower() == "none":
        successFlag = None
    if failureFlag and failureFlag.lower() == "none":
        failureFlag = None

    directory_script_runner(
        log=log,
        pathToScriptDirectory=pathToDirectory,
        databaseName=databaseName,
        loginPath=loginPath,
        successRule=successFlag,
        failureRule=failureFlag
    )

    return

if __name__ == '__main__':
    main()
